from unittest import TestCase
from uuid import uuid5, NAMESPACE_DNS
from urllib.parse import urlencode

from django.conf import settings
from django.core.cache import caches
from django.http.response import HttpResponse
from django.test import RequestFactory
from django.views.generic import View

from .middleware import DropDuplicatedRequests


CACHE_NAME = getattr(settings, "DUPLICATED_REQUESTS_CACHE_NAME", "default")
COOKIE_NAME = getattr(settings, "DUPLICATED_REQUESTS_COOKIE_NAME", "dj-request-id")


class TestDropDuplicatedRequests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.middleware = DropDuplicatedRequests()

    def tearDown(self):
        cache = caches[CACHE_NAME]
        cache.clear()

    def _call_view_using_middleware(self, method, set_cookie=True, path="/", body={}):
        class TestView(View):
            def get(self, request):
                return HttpResponse()

            put = post = patch = delete = get

        # Get a new request and process it using middleware
        request = getattr(self.factory, method)(path, body)
        if set_cookie:
            request.COOKIES[COOKIE_NAME] = "not-so-unique-id"
        response = self.middleware.process_request(request)

        if response is None:
            response = TestView.as_view()(request)
        return self.middleware.process_response(request, response)

    def test_double_get(self):
        response_1 = self._call_view_using_middleware("get")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("get")
        self.assertEqual(response_2.status_code, 200)

    def test_double_post(self):
        response_1 = self._call_view_using_middleware("post", body={"a": "a"})
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("post", body={"a": "a"})
        self.assertEqual(response_2.status_code, 304)

    def test_double_post_without_cookie(self):
        response_1 = self._call_view_using_middleware("post", False)
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("post", False)
        self.assertEqual(response_2.status_code, 200)

    def test_double_put(self):
        response_1 = self._call_view_using_middleware("put")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put")
        self.assertEqual(response_2.status_code, 304)

    def test_double_patch(self):
        response_1 = self._call_view_using_middleware("patch")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("patch")
        self.assertEqual(response_2.status_code, 304)

    def test_double_delete(self):
        response_1 = self._call_view_using_middleware("delete")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("delete")
        self.assertEqual(response_2.status_code, 304)

    def test_double_requests_different_method(self):
        response_1 = self._call_view_using_middleware("patch")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put")
        self.assertEqual(response_2.status_code, 200)

    def test_double_requests_different_path(self):
        response_1 = self._call_view_using_middleware("put", path="/123")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put", path="/456")
        self.assertEqual(response_2.status_code, 200)

    def test_double_requests_same_path(self):
        response_1 = self._call_view_using_middleware("put", path="/123")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put", path="/123")
        self.assertEqual(response_2.status_code, 304)

    def test_double_requests_different_get_params(self):
        response_1 = self._call_view_using_middleware("put", path="/?a=123")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put", path="/?a=456")
        self.assertEqual(response_2.status_code, 200)

    def test_double_requests_same_get_params(self):
        response_1 = self._call_view_using_middleware("put", path="/?a=123")
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put", path="/?a=123")
        self.assertEqual(response_2.status_code, 304)

    def test_double_requests_different_body(self):
        response_1 = self._call_view_using_middleware("put", body={"a": "b"})
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put", body={"a": "c"})
        self.assertEqual(response_2.status_code, 200)

    def test_double_requests_same_body(self):
        response_1 = self._call_view_using_middleware("put", body={"a": "b"})
        self.assertEqual(response_1.status_code, 200)
        response_2 = self._call_view_using_middleware("put", body={"a": "b"})
        self.assertEqual(response_2.status_code, 304)

    def test_set_cookie(self):
        response = self._call_view_using_middleware('get')
        self.assertIn(COOKIE_NAME, response.cookies)



from uuid import uuid4, uuid5, NAMESPACE_DNS
from urllib.parse import urlencode

from django.conf import settings
from django.core.cache import caches
from django.http.response import HttpResponseNotModified


try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:  # pragma: nocover
    MiddlewareMixin = object

CACHE_NAME = getattr(settings, "DUPLICATED_REQUESTS_CACHE_NAME", "default")
CACHE_TIMEOUT = getattr(settings, "DUPLICATED_REQUESTS_CACHE_TIMEOUT", 5)
COOKIE_NAME = getattr(settings, "DUPLICATED_REQUESTS_COOKIE_NAME", "dj-request-id")
COOKIE_PREFIX = getattr(settings, "DUPLICATED_REQUESTS_COOKIE_PREFIX", "request-id-")


class DropDuplicatedRequests(MiddlewareMixin):
    """
    Middleware that drops requests made in quick succession.
    Uses Django's caching system to check/save each request.
    """

    def _get_request_hash(self, request):
        """
        Generates a unique key based on request path, method, body and arguments
        """
        hash_value = uuid5(
            NAMESPACE_DNS,
            request.path_info
            + "--"
            + request.method.lower()
            + "--"
            + urlencode(request.GET)
            + "--"
            + request.body.decode("utf-8"),
        ).node
        return str(hash_value)

    def process_request(self, request):
        """
        Stores a unique key per request in the cache, if it already exists, returns 304
        """
        if not request.method.lower() in ("post", "put", "delete", "patch"):
            return

        cookie_value = request.COOKIES.get(COOKIE_NAME)
        if not cookie_value:
            return

        cache_key = cookie_value + self._get_request_hash(request)

        cache = caches[CACHE_NAME]
        if cache_key in cache:
            return HttpResponseNotModified()
        cache.set(cache_key, True, CACHE_TIMEOUT)

    def process_response(self, request, response):
        """
        Sends a cookie with a unique hash to identify requests that are the same
        but from different sources
        """
        response.set_cookie(COOKIE_NAME, COOKIE_PREFIX + uuid4().hex)
        return response
